"""Document chat bots."""
from pathlib import Path

import typer

from llamabot import QueryBot

from .utils import exit_if_asked, uniform_prompt

app = typer.Typer()


@app.command()
def chat(
    model_name: str = typer.Option(
        "mistral/mistral-medium", help="Name of the LLM to use."
    ),
    initial_message: str = typer.Option(..., help="Initial message for the bot."),
    panel: bool = typer.Option(True, help="Whether to use Panel or not."),
    doc_path: Path = typer.Argument(
        "", help="Path to the document you wish to chat with."
    ),
):
    """Chat with your document.

    :param model_name: Name of the LLM to use.
    :param panel: Whether to use Panel or not. If not, we default to using CLI chat.
    :param initial_message: The initial message to send to the user.
    :param doc_path: Path to the document you wish to chat with.
    """
    stream_target = "stdout"
    if panel:
        stream_target = "panel"

    bot = QueryBot(
        system_prompt=(
            "You are a bot that can answer questions about a document provided to you. "
            "Based on the document, respond to the query that is given to you. "
            "As much as possible, quote from the context that is provided to you."
        ),
        collection_name=doc_path.stem.lower().replace(" ", "-"),
        document_paths=[doc_path],
        model_name=model_name,
        initial_message=initial_message,
        stream_target=stream_target,
    )
    typer.echo(
        (
            "I've embedded your document. "
            "Ask me anything! "
            "Otherwise, type 'exit' or 'quit' at anytime to exit."
        )
    )

    if panel:
        print("Serving your document in a panel...")
        bot.serve()

    else:
        while True:
            query = uniform_prompt()
            exit_if_asked(query)
            bot(query)
            typer.echo("\n\n")



"""Tests for the prompt decorator."""
import pytest
from llamabot.prompt_manager import prompt


def test_prompt_no_args():
    """Test that the decorated prompt function returns the correct string."""

    @prompt
    def test_func():
        """This is a test function."""

    assert test_func() == "This is a test function."


def test_prompt_with_args():
    """Test that the decorated prompt function returns the correct string with args."""

    @prompt
    def test_func(a, b):
        """This function takes two arguments: {{a}} and {{b}}. # noqa: DAR101"""

    assert (
        test_func(1, 2) == "This function takes two arguments: 1 and 2. # noqa: DAR101"
    )


def test_prompt_with_kwargs():
    """Test that the decorated prompt function
    returns the correct string with kwargs."""

    @prompt
    def test_func(a=1, b=2):
        """This function takes two arguments: {{a}} and {{b}}. # noqa: DAR101"""

    assert (
        test_func(a=3, b=4)
        == "This function takes two arguments: 3 and 4. # noqa: DAR101"
    )


def test_prompt_with_args_and_kwargs():
    """Test that the decorated prompt function
    returns the correct string with args and kwargs."""

    @prompt
    def test_func(a, b=2):
        """This function takes two arguments: {{a}} and {{b}}. # noqa: DAR101"""

    assert (
        test_func(1, b=3)
        == "This function takes two arguments: 1 and 3. # noqa: DAR101"
    )


def test_prompt_with_missing_kwargs():
    """Test that the decorated prompt function
    raises an error when kwargs are missing."""

    @prompt
    def test_func(a=1, b=2):
        """This function takes two arguments: {{a}} and {{b}}. # noqa: DAR101"""

    with pytest.raises(ValueError):
        test_func(a=3)



"""Class definition for QueryBot."""
import contextvars
from pathlib import Path
from typing import Optional
from dotenv import load_dotenv
from loguru import logger

from llamabot.config import default_language_model
from llamabot.bot.simplebot import SimpleBot
from llamabot.components.messages import AIMessage, HumanMessage
from llamabot.components.docstore import LanceDBDocStore
from llamabot.components.chatui import ChatUIMixin
from llamabot.components.messages import (
    RetrievedMessage,
)
from slugify import slugify

load_dotenv()


CACHE_DIR = Path.home() / ".llamabot" / "cache"
prompt_recorder_var = contextvars.ContextVar("prompt_recorder")


class QueryBot(SimpleBot, ChatUIMixin):
    """QueryBot is a bot that uses the DocumentStore to answer questions about a document."""

    def __init__(
        self,
        system_prompt: str,
        collection_name: str,
        initial_message: Optional[str] = None,
        document_paths: Optional[Path | list[Path]] = None,
        temperature: float = 0.0,
        model_name: str = default_language_model(),
        stream_target: str = "stdout",
        **kwargs,
    ):
        SimpleBot.__init__(
            self,
            system_prompt=system_prompt,
            temperature=temperature,
            model_name=model_name,
            stream_target=stream_target,
            **kwargs,
        )
        logger.debug("Initializing LanceDB DocStore...")
        self.lancedb_store = LanceDBDocStore(table_name=slugify(collection_name))
        self.lancedb_store.reset()
        if document_paths:
            logger.debug("Adding documents to LanceDB DocStore...")
            self.lancedb_store.add_documents(document_paths=document_paths)

        self.response_budget = 2_000

        ChatUIMixin.__init__(self, initial_message)

    def __call__(self, query: str, n_results: int = 10) -> AIMessage:
        """Query documents within QueryBot's document store.

        We use RAG to query out documents.

        :param query: The query to make of the documents.
        """
        messages = []

        # context_budget = model_context_window_sizes.get(
        #     self.model_name, DEFAULT_TOKEN_BUDGET
        # )

        retreived_messages = set()
        logger.debug(f"Retrieving {n_results} documents from LanceDB DocStore...")
        retrieved_messages = retreived_messages.union(
            self.lancedb_store.retrieve(query, n_results)
        )
        retrieved = [RetrievedMessage(content=chunk) for chunk in retrieved_messages]
        messages.extend(retrieved)
        messages.append(HumanMessage(content=query))
        if self.stream_target == "stdout":
            response: AIMessage = self.stream_stdout(messages)
            return response
        elif self.stream_target == "panel":
            return self.stream_panel(messages)



"""
This module provides a set of tests for the PromptRecorder
and autorecord functions in the llamabot.recorder module.
"""
import pandas as pd

from llamabot.recorder import PromptRecorder, autorecord


def test_init():
    """Test that the PromptRecorder is initialized correctly."""
    recorder = PromptRecorder()
    assert recorder.prompts_and_responses == []


def test_log():
    """Test that the PromptRecorder logs prompts and responses correctly."""
    recorder = PromptRecorder()
    recorder.log("prompt1", "response1")
    assert recorder.prompts_and_responses == [
        {"prompt": "prompt1", "response": "response1"}
    ]


def test_repr():
    """Test that the PromptRecorder is represented correctly."""
    recorder = PromptRecorder()
    recorder.log("prompt1", "response1")
    assert isinstance(recorder.__repr__(), str)


def test_dataframe():
    """Test that the PromptRecorder is represented correctly as a dataframe."""
    recorder = PromptRecorder()
    recorder.log("prompt1", "response1")
    assert isinstance(recorder.dataframe(), pd.DataFrame)


def test_autorecord_no_context():
    """Test that autorecord works without a context."""
    prompt = "Hello"
    response = "Hi"
    autorecord(prompt, response)


def test_autorecord_with_context():
    """Test that autorecord works with a context."""
    prompt = "Hello"
    response = "Hi"
    recorder = PromptRecorder()
    with recorder:
        autorecord(prompt, response)
    assert len(recorder.prompts_and_responses) == 1
    assert recorder.prompts_and_responses[0]["prompt"] == prompt
    assert recorder.prompts_and_responses[0]["response"] == response


def test_autorecord_multiple_with_context():
    """Test that autorecord works with a context multiple times."""
    prompts = ["Hello", "How are you?"]
    responses = ["Hi", "I'm good, thanks!"]
    recorder = PromptRecorder()
    with recorder:
        for p, r in zip(prompts, responses):
            autorecord(p, r)
    assert len(recorder.prompts_and_responses) == 2
    for i, (p, r) in enumerate(zip(prompts, responses)):
        assert recorder.prompts_and_responses[i]["prompt"] == p
        assert recorder.prompts_and_responses[i]["response"] == r



"""Paths to configuration files and directories."""
import os
from pathlib import Path

from dotenv import load_dotenv

llamabotrc_path = Path.home() / ".llamabot/.llamabotrc"

llamabot_config_dir = Path.home() / ".llamabot"


def default_language_model():
    """Return the default language model to be used.

    :return: The default language model to be used.
    """

    load_dotenv(llamabotrc_path)

    return os.getenv("DEFAULT_LANGUAGE_MODEL", "mistral/mistral-medium")



"""CLI for chatting with a code repository."""
from typing import List, Optional
import typer
import git
import tempfile
from pathlib import Path
from llamabot import QueryBot
from .utils import exit_if_asked, uniform_prompt
from slugify import slugify
from loguru import logger

app = typer.Typer()


@app.command()
def chat(
    repo_url: str,
    checkout: str = "main",
    source_file_extensions: List[str] = [
        "py",
        "jl",
        "R",
        "ipynb",
        "md",
        "tex",
        "txt",
        "lr",
        "rst",
    ],
    model_name: str = "gpt-4-0125-preview",
    initial_message: Optional[str] = None,
    panel: bool = True,
):
    """Chat with a code repository."""
    # Create a temporary directory
    logger.info("Creating temporary directory...")
    temp_dir = tempfile.TemporaryDirectory(dir="/tmp")

    # Clone the repository into the temporary directory
    logger.info("Cloning repository...")
    repo = git.Repo.clone_from(repo_url, temp_dir.name)

    # checkout the specified branch or tag (i.e. "checkout")
    logger.info(f"Checking out branch {checkout}...")
    repo.git.checkout(checkout)

    # Set the root directory to the cloned repository
    root_dir = Path(temp_dir.name)

    source_files = []
    for extension in source_file_extensions:
        files = list(root_dir.rglob(f"*.{extension}"))
        print(f"Found {len(files)} files with extension {extension}.")
        source_files.extend(files)

    stream_target = "stdout"
    if panel:
        stream_target = "panel"

    bot = QueryBot(
        system_prompt="You are a knowledgeable git repository author. Your answers come from the repository. If the answer is not in the repository, say 'I don't know'.",
        collection_name=slugify(repo_url + ":" + checkout),
        document_paths=source_files,
        model_name=model_name,
        initial_message=initial_message,
        stream_target=stream_target,
    )
    if panel:
        print("Serving your document in a panel...")
        bot.serve()

    else:
        while True:
            query = uniform_prompt()
            exit_if_asked(query)
            bot(query)
            typer.echo("\n\n")



"""Constants for model tokens and token budgets."""
DEFAULT_TOKEN_BUDGET = 4_096

# Reference: https://platform.openai.com/docs/models
# For the Ollama models, refer to individual model documentation page: https://ollama.ai/library
# Please update the numbers here as necessary.
model_context_window_sizes = {
    "gpt-3.5-turbo-0301": 4_097,
    "gpt-3.5-turbo-0613": 4_097,
    "gpt-3.5-turbo-16k-0613": 16_385,
    "gpt-3.5-turbo-16k": 16_385,
    "gpt-3.5-turbo-instruct-0914": 4_097,
    "gpt-3.5-turbo-instruct": 4_097,
    "gpt-3.5-turbo": 4_097,
    "gpt-4-0314": 8_192,
    "gpt-4-0613": 8_192,
    "gpt-4-32k": 32_768,
    "gpt-4-32k-0613": 32_768,
    "gpt-4": 8_192,
    "llama2": 4_096,
    "vicuna:7b-16k": 16_385,
    "mistral/mistral-tiny": 4_096,
    "mistral/mistral-small": 8_192,
    "mistral/mistral-medium": 32_768,
}



"""Blogging prompts."""
from pydantic import BaseModel

from llamabot import SimpleBot
from llamabot.prompt_manager import prompt


class BlogInformation(BaseModel):
    """Blog information."""

    tags: list[str]
    summary: str
    title: str


def blogging_bot():
    """Blogging bot.

    :return: The blogging bot.
    """
    bot = SimpleBot(
        "You are an expert blogger. Whenever you use hashtags, "
        "they are always lowercase.",
        temperature=0.3,
    )
    return bot


@prompt
def blog_title_tags_summary(blog_post, blog_info_model):
    """This is a blog post that I just wrote.

    {{ blog_post }}

    Please return for me up to 15 blog tags in lowercase.
    They should be at most 2 words long.
    Return one tag per line.
    No symbols, just the words, all lowercase.

    Also, please return for me a summary of the blog post,
    written in first-person tone,
    that is at most 100 words long.
    Use emojis where appropriate.

    Finally, please suggest a title for the blog post.
    It should be entertaining without being overly so,
    and should entice readers to read the blog post.

    Respond with a JSON formatted as follows:

    {{ blog_info_model | schema }}

    # noqa: DAR101
    """


@prompt
def compose_linkedin_post(blog_post):
    """This is a blog post that I just wrote:

    {{ blog_post }}

    Please compose for me a LinkedIn post
    that entices my network on LinkedIn to read it.
    Ensure that there is a call to action to interact with the post after reading
    to react with it, comment on it, or share the post with others,
    and to support my work on Patreon.
    My Patreon link is https://patreon.com/ericmjl/
    Include hashtags inline with the LinkedIn post and at the end of the post too.
    Please return this for me in JSON format using the following schema:

    {
        "post_text": <post_text>
    }

    #noqa: DAR101
    """


@prompt
def compose_patreon_post(blog_post):
    """This is a blog post that I just wrote:

    {{ blog_post }}

    Please compose for me a Patreon post
    that entices my patrons on Patreon to read it.
    Ensure that there is a call to action to interact with the post after reading it,
    such as asking a question, or suggesting ideas that I did not write about.
    Please use emojis where appropriate as well!
    Based on the content of the blog post, I also need a title for the Patreon post,
    a 140 character text preview, and tags for the post.
    Please return this for me in JSON format using the following schema:

    {
        "title": <title>,
        "text_preview": <text_preview>,
        "post_text": <post_text>,
        "tags": [<tag1>, <tag2>, ...],
    }

    #noqa: DAR101
    """


@prompt
def compose_twitter_post(blog_post):
    """This is a blog post that I just wrote:

    {{ blog_post }}

    Please compose for me a Twitter post
    that entices my followers on Twitter to read it.
    Ensure that there is a call to action to interact with the post after reading it,
    such as retweeting, commenting, or sharing it with others,
    and to support my work on Patreon.
    My Patreon link is https://patreon.com/ericmjl/
    Include hashtags inline with the Twitter post.

    Please return this for me in JSON format using the following schema:

    {
        "post_text": <post_text>
    }

    #noqa: DAR101
    """



"""Decorators for turning Python functions into Jinja2-templated prompts.

This module contains decorators for turning Python functions' docstrings
into Jinja2-templated functions that accept function arguments
and return the Jinja2 templated docstrings as strings when called.

Inspired from the Outlines library.
"""


from functools import wraps
import jinja2
from jinja2 import meta
import inspect


def prompt(func):
    """Wrap a Python function into a Jinja2-templated prompt.

    :param func: The function to wrap.
    :return: The wrapped function.
    """

    @wraps(func)
    def wrapper(*args, **kwargs) -> str:
        """Wrapper function.

        :param args: Positional arguments to the function.
        :param kwargs: Keyword arguments to the function.
        :return: The Jinja2-templated docstring.
        :raises ValueError: If a variable in the docstring
            is not passed into the function.
        """
        # get the function's docstring.
        docstring = func.__doc__

        # map args and kwargs onto func's signature.
        signature = inspect.signature(func)
        kwargs = signature.bind(*args, **kwargs).arguments

        # create a Jinja2 environment
        env = jinja2.Environment()

        # parse the docstring
        parsed_content = env.parse(docstring)

        # get all variables in the docstring
        variables = meta.find_undeclared_variables(parsed_content)

        # check if all variables are in kwargs
        for var in variables:
            if var not in kwargs:
                raise ValueError(f"Variable '{var}' was not passed into the function")

        # interpolate docstring with args and kwargs
        template = jinja2.Template(docstring)
        return template.render(**kwargs)

    return wrapper



"""Serve up LlamaBots as a FastAPI endpoint."""
from pathlib import Path
import typer
from typing import List
from fastapi import FastAPI
from llamabot.bot import QueryBot
import uvicorn

api = FastAPI()
cli = typer.Typer()


@cli.command()
def querybot(
    system_prompt: str = typer.Option(..., help="System prompt."),
    collection_name: str = typer.Option(..., help="Name of the collection."),
    document_paths: List[Path] = typer.Option(..., help="Paths to the documents."),
    model_name: str = typer.Option(
        "mistral/mistral-medium", help="Name of the model to use."
    ),
    host: str = typer.Option("0.0.0.0", help="Host to serve the API on."),
    port: int = typer.Option(6363, help="Port to serve the API on."),
):
    """Serve up a LlamaBot as a FastAPI endpoint."""
    bot = QueryBot(
        system_prompt=system_prompt,
        collection_name=collection_name,
        document_paths=document_paths,
        model_name=model_name,
    )

    api.add_api_route(
        "/",
        bot.create_endpoint(),
    )

    uvicorn.run(api, host=host, port=port)



"""Zotero prompt library."""
from llamabot.prompt_manager import prompt


@prompt
def retrieverbot_sysprompt():
    """You are an expert in retrieving information from JSON files."""


@prompt
def docbot_sysprompt():
    """You are an expert in answering questions about any paper.

    Respond the way that Richard Feynman would answer questions
    to a 2nd year undergraduate student.
    """


@prompt
def paper_summary():
    """Please synthesize a summary of this paper using the information provided to you.

    Your summary should not be a mere regurgitation of the abstract.
    Rather, your summary should highlight the key findings,
    methodology, and implications.
    """


@prompt
def get_key(query: str = ""):
    """Based on the context provided,
    return for me all of the keys that were provided for the query.

    Ensure that you're using the titles and abstracts.

    {{ query }}

    Return this as JSON:

    {
        "key": [<key1>, <key2>, ...]
    }

    The list should include only the keys for the papers that are relevant to the query.

    Do not return anything else except JSON.
    Ensure that you do not return any explanations of your response.
    If you cannot help with the request, return an empty list.

    # noqa: DAR101
    """



"""Knowledge Graph bot."""
import json
from llamabot.components.messages import HumanMessage
from llamabot.config import default_language_model

from llamabot.prompt_manager import prompt
from .simplebot import SimpleBot


@prompt
def kgbot_sysprompt() -> str:
    """You are an expert ontologist. You are tasked with taking in a chunk of text
    and extrapolating as many relationships as possible from that text.

    For each relationship, return a JSON according to the following schema:

    {
        "subject": "string",
        "predicate": "string",
        "object": "string"
    }
    """


class KGBot(SimpleBot):
    """KGBot is the Knowledge Graph bot.

    Tested with mistral-medium.

    It takes in a chunk of text and returns a JSON of triplets.
    """

    def __init__(self):
        SimpleBot.__init__(
            self, system_prompt=kgbot_sysprompt(), model_name=default_language_model()
        )

    def __call__(self, query: str) -> dict:
        """Call the bot with a query and return a JSON of triplets.

        :param query: The query to use.
        """
        response = self.generate_response([HumanMessage(content=query)])
        return json.loads(response.content)



"""Bot abstractions that let me quickly build new GPT-based applications."""


import panel as pn
from dotenv import load_dotenv

from llamabot.config import llamabotrc_path

from .chatbot import ChatBot
from .querybot import QueryBot
from .simplebot import SimpleBot
from .imagebot import ImageBot

pn.extension()
load_dotenv()

if llamabotrc_path.exists():
    load_dotenv(llamabotrc_path)


__all__ = ["SimpleBot", "ChatBot", "QueryBot", "ImageBot"]



"""Run llamabot codebot panel apps."""
import os

import typer

from llamabot.prompt_library import python

app = typer.Typer()


@app.command()
def codebot(port: int = 5050, address: str = "0.0.0.0"):
    """Run the codebot app.

    :param port: Port to run the app on.
    :param address: Address to run the app on.
    """

    codebot_app = python.create_panel_app()
    os.environ["BOKEH_ALLOW_WS_ORIGIN"] = "*"
    codebot_app.show(title="Codebot", port=port, address=address, open=False)


if __name__ == "__main__":
    app()



"""ImageBot module for generating images."""

from openai import OpenAI
from IPython.display import display, Image
import requests
from pathlib import Path
from typing import Optional, Union
from llamabot.components.messages import AIMessage


class ImageBot:
    """ImageBot class for generating images.

    :param model: The model to use. Defaults to "dall-e-3".
    :param size: The size of the image to generate. Defaults to "1024x1024".
    :param quality: The quality of the image to generate. Defaults to "standard".
    :param n: The number of images to generate. Defaults to 1.
    """

    def __init__(self, model="dall-e-3", size="1024x1024", quality="hd", n=1):
        self.client = OpenAI()
        self.model = model
        self.size = size
        self.quality = quality
        self.n = n

    def __call__(
        self, prompt: str, save_path: Optional[Path] = None
    ) -> Union[str, Path]:
        """Generate an image from a prompt.

        :param prompt: The prompt to generate an image from.
        :param save_path: The path to save the generated image to.
            If it is empty, then we will generate a filename from the prompt.
        :return: The URL of the generated image if running in a Jupyter notebook (str),
            otherwise a pathlib.Path object pointing to the generated image.
        :raises Exception: If no image URL is found in the response.
        """
        response = self.client.images.generate(
            model=self.model,
            prompt=prompt,
            size=self.size,
            quality=self.quality,
            n=self.n,
        )
        image_url = response.data[0].url
        if not image_url:
            raise Exception("No image URL found in response! Please try again.")

        # Check if running in a Jupyter notebook
        if is_running_in_jupyter():
            display(Image(url=image_url))
            return image_url

        image_data = requests.get(image_url).content

        if not save_path:
            filename = filename_bot(prompt).content
            save_path = Path(f"{filename}.jpg")
        with open(save_path, "wb") as file:
            file.write(image_data)
        return save_path


def is_running_in_jupyter() -> bool:
    """Check if running in a Jupyter notebook.

    :return: True if running in a Jupyter notebook, otherwise False.
    """
    try:
        get_ipython()
        return True
    except NameError:
        return False


def filename_bot(image_prompt: str) -> AIMessage:
    """Generate a filename from an image prompt.

    :param image_prompt: The image prompt to generate a filename from.
    :return: The filename generated from the image prompt within an AIMessage object.
    """
    from llamabot import SimpleBot

    bot = SimpleBot(
        "You are a helpful filenaming assistant. "
        "Filenames should use underscores instead of spaces, "
        "and should be all lowercase. "
        "Exclude the file extension. "
        "Give me a compact filename for the following prompt:"
    )
    response = bot(image_prompt)
    return response



"""Top-level command line interface for llamabot."""

import os
from datetime import datetime
from pathlib import Path

import typer

from llamabot import ChatBot, PromptRecorder

from . import blog, configure, doc, git, python, tutorial, zotero, repo, serve
from .utils import exit_if_asked, uniform_prompt

app = typer.Typer()
app.add_typer(
    python.app,
    name="python",
    help="Python bot for generating docstrings, code, and tests.",
)
app.add_typer(
    git.gitapp,
    name="git",
    help="Automatic commit message generation and pre-commit hook installation.",
)
app.add_typer(
    tutorial.app,
    name="tutorial",
    help="Automatically generate tutorials for given source files.",
)
app.add_typer(zotero.app, name="zotero", help="Chat with your Zotero library.")
app.add_typer(doc.app, name="doc", help="Chat with your documents.")
app.add_typer(
    blog.app,
    name="blog",
    help=(
        "Summarize blog posts, generate social media posts, "
        "and apply SEMBR to blog posts in a chat UI."
    ),
)
app.add_typer(configure.app, name="configure", help="Configure LlamaBot.")
app.add_typer(repo.app, name="repo", help="Chat with a code repository.")
app.add_typer(
    serve.cli, name="serve", help="Serve up a LlamaBot as a FastAPI endpoint."
)


@app.command()
def version():
    """
    Print the version of llamabot.
    """
    from llamabot.version import version

    typer.echo(version)


@app.command()
def chat(save: bool = typer.Option(True, help="Whether to save the chat to a file.")):
    """Chat with LlamaBot's ChatBot.

    :param save: Whether to save the chat to a file.
    """
    pr = PromptRecorder()

    bot = ChatBot(
        "You are a chatbot. Respond to the user. "
        "Ensure your responses are Markdown compatible.",
        session_name=f"{datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}-chat",
    )

    # Save chat to file
    save_filename = f"{datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}-chat.md"

    while True:
        with pr:
            input = uniform_prompt()
            exit_if_asked(input)
            bot(input)
            typer.echo("\n\n")

            if save:
                pr.save(Path(save_filename))


@app.command()
def clear_cache():
    """Clear the Llamabot cache."""
    CACHE_DIR = Path.home() / ".llamabot" / "cache"
    os.system("rm -rf {}".format(CACHE_DIR))


if __name__ == "__main__":
    app()



"""Zotero library wrappers."""
import json
from dataclasses import dataclass, field
from pathlib import Path

from pyzotero.zotero import Zotero
from rich.progress import Progress, SpinnerColumn, TextColumn

from .utils import load_zotero

progress = Progress(
    SpinnerColumn(),
    TextColumn("[progress.description]{task.description}"),
    transient=False,
)


@dataclass
class ZoteroLibrary:
    """Zotero library object.

    Stores a list of Zotero items.
    """

    zot: Zotero = field(default_factory=load_zotero)
    json_dir: Path = field(default=None)
    articles_only: bool = False

    def __post_init__(
        self,
    ):
        """Post-initialization hook

        If json_dir is set, load the library from the JSON files in that directory
        and skip querying zotero for everything.
        """
        if self.json_dir is not None:
            # Load the library from the JSON files.
            items = []
            for json_file in self.json_dir.glob("*.json"):
                with json_file.open("r+") as f:
                    items.append(json.loads(f.read()))
        else:
            with progress:
                task = progress.add_task("Synchronizing your Zotero library...")
                items = self.zot.everything(self.zot.items())
                progress.remove_task(task)

        condition = lambda i: True  # noqa: E731
        if self.articles_only:
            condition = lambda i: i.has_pdf()  # noqa: E731
        library = [ZoteroItem(i, library=self) for i in items]
        # Filter for the condition using the ZoteroItem.has_pdf() method.
        library = [i for i in library if condition(i)]
        self.library = {i["key"]: i for i in library}

    def __getitem__(self, key):
        """Get item by key.

        :param key: Key to get.
        :return: Value for key.
        """
        return self.library[key]

    def keys(self):
        """Return all of the keys from the library.

        :return: A list of keys.
        """
        return [i for i in self.library]

    def to_json(self, dir: Path, has_pdf=True):
        """Save the library to a JSON file.

        :param dir: Directory in which to save the JSONs.
        :param has_pdf: Whether to save only items with PDFs.
        """
        for key, item in self.library.items():
            if has_pdf and item.has_pdf():
                with open((dir / f"{key}.json"), "w+") as f:
                    f.write(json.dumps(item.info) + "\n")

    def key_title_map(self, inverse=False) -> dict[str, str]:
        """Return the paper titles from the library.

        :param inverse: Whether to invert the mapping.
        :return: A list of paper titles.
        """
        mapping = {}
        for k, v in self.library.items():
            data = v.info["data"]
            title = data["title"] if "title" in data.keys() else None
            if title is None:
                continue
            title = v.info["data"]["title"]
            mapping[k] = title

        if inverse:
            return {v: k for k, v in mapping.items()}
        return mapping


@dataclass
class ZoteroItem:
    """Zotero item."""

    info: dict
    library: ZoteroLibrary

    def __getitem__(self, key):
        """Get item by key.

        Allows for accessing nested keys via a "."-delimited string.

        :param key: Key to get.
        :return: Value for key.
        :raises KeyError: If key is not found.
        """
        # Key should be a string that is dot-delimited.
        # we split the string into a list of keys
        # Then we access the keys in order.

        keys = key.split(".")
        value = self.info
        for k in keys:
            try:
                value = value[k]
            except KeyError:
                raise KeyError(f"Key {k} not found in {value}.")
        return value

    def has_pdf(self):
        """Check if this item has a PDF or not.

        We check the "links" section, which typically looks like this:

        ```json
        {
            'self': {
                'href': 'https://api.zotero.org/users/5334442/items/K3WYABBQ',
                'type': 'application/json'
            },
            'alternate': {
                'href': 'https://www.zotero.org/ericmjl/items/K3WYABBQ',
                'type': 'text/html'
            },
            'attachment': {
                'href': 'https://api.zotero.org/users/5334442/items/U6X244QK',
                'type': 'application/json',
                'attachmentType': 'application/pdf',
                'attachmentSize': 4351619
            }
        }
        ```

        :return: True if this item has a PDF, False otherwise.
        """
        if self.get("links.attachment.attachmentType") == "application/pdf":
            return True
        return False

    def get(self, key_string, default_value=None):
        try:
            return self[key_string]
        except KeyError:
            return default_value

    def pdf(self):
        """Get the PDF entry for this item.

        :return: PDF entry.
        :raises KeyError: If no PDF is found.
        """
        if self.has_pdf():
            return self["links.attachment"]
        raise KeyError("No PDF found.")

    def download_pdf(self, directory: Path) -> Path:
        """Download the PDF for this item.

        :param directory: Directory to download the PDF to.
        :return: Path to the downloaded PDF.
        """
        pdf = self.pdf()
        key = pdf["href"].split("/")[-1]
        fpath = directory / f"{key}.pdf"
        with fpath.open("wb") as f:
            f.write(self.library.zot.file(key))
        return fpath

    def download_abstract(self, directory: Path) -> Path:
        """Download the abstract for this item.

        :param directory: Directory to download the abstract to.
        :return: Path to the downloaded abstract.
        """
        fpath = directory / "abstract.txt"
        with fpath.open("w+") as f:
            f.write(self["data"]["abstractNote"])
        return fpath



"""Top-level API for llamabot.

This is the file from which you can do:

    from llamabot import some_function

Use it to control the top-level API of your Python data science project.
"""
# Ensure that ~/.llamabotrc exists.
from pathlib import Path

from .bot import ChatBot, QueryBot, SimpleBot, ImageBot
from .recorder import PromptRecorder

__all__ = ["ChatBot", "ImageBot", "SimpleBot", "QueryBot", "PromptRecorder"]


(Path.home() / ".llamabot").mkdir(parents=True, exist_ok=True)



"""A component that provides tools to the chatbot."""
import inspect
from openai.types.chat import ChatCompletionMessageToolCall
import typing
import json


def python_type_to_json_type(python_type):
    """Map Python types to JSON Schema types."""
    type_map = {
        str: "string",
        int: "integer",
        float: "number",
        bool: "boolean",
        list: "array",
        dict: "object",
        # Add more mappings as needed
    }
    return type_map.get(python_type, "any")


def type_to_str(type_hint):
    """Convert type hints to JSON-friendly string representations."""
    if type_hint == inspect.Parameter.empty:
        return "any"
    if getattr(type_hint, "__origin__", None) is typing.Literal:
        # Handling typing.Literal to convert it to JSON enum format
        return {"enum": list(type_hint.__args__)}
    if hasattr(type_hint, "__origin__"):  # For handling generic types like List[str]
        origin = type_hint.__origin__
        if origin is list:
            # Assuming only simple types like List[str], not nested like List[List[str]]
            args = type_hint.__args__[0]
            return f"array of {python_type_to_json_type(args)}"
        # Handle other generic types (like Dict, Tuple) here as needed
    return python_type_to_json_type(type_hint)


def describe_function(func):
    """Describe a function as a JSON Schema.

    :param func: The function to describe.
    :return: A JSON Schema describing the function.
    """
    # Extract the signature of the function
    signature = inspect.signature(func)
    docstring = inspect.getdoc(func)

    # Assuming the first line of the docstring is the function description
    function_description = docstring.split("\n")[0]

    # Extracting parameter information
    parameters = {}
    for name, param in signature.parameters.items():
        # Assume the description is in the format: `name: description`
        param_description = [
            line.split(": ")[1]
            for line in docstring.split("\n")
            if line.startswith(name + ":")
        ]
        param_description = param_description[0] if param_description else ""

        # Building the parameter info
        param_type = type_to_str(param.annotation)
        param_info = {"description": param_description}
        if isinstance(param_type, dict):
            # If the type is a dictionary (e.g., for enum), merge it with param_info
            param_info.update(param_type)
        else:
            param_info["type"] = param_type

        parameters[name] = param_info

    # Required parameters are those without default values
    required_params = [
        name
        for name, param in signature.parameters.items()
        if param.default == param.empty
    ]

    # Constructing the final description
    result = {
        "type": "function",
        "function": {
            "name": func.__name__,
            "description": function_description,
            "parameters": {
                "type": "object",
                "properties": parameters,
                "required": required_params,
            },
        },
    }

    return result


class Tools:
    """A component that provides tools to the chatbot."""

    def __init__(self, *functions):
        self._available_tools = {func.__name__: func for func in functions}
        self._schemas = {func.__name__: describe_function(func) for func in functions}

    def __call__(
        self, tool_calls: list[ChatCompletionMessageToolCall]
    ) -> dict[str, any]:
        """Call on the tools that were provided.

        :param tool_name: The name of the tool to call.
        :param kwargs: The arguments to pass to the tool.
        """
        result = {}
        if tool_calls:
            for tool_call in tool_calls:
                func_name = tool_call.function.name
                func = self._available_tools[func_name]
                func_kwargs = json.loads(tool_call.function.arguments)
                result[func_name] = func(**func_kwargs)
        return result

    def schemas(self):
        """Return the schemas for the tools."""
        return list(self._schemas.values())



"""LlamaBot configuration."""
from openai import OpenAI

import typer
from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter
from .utils import configure_environment_variable


app = typer.Typer()


@app.command()
def default_model(model_name=None):
    """Configure the default model for llamabot.

    If no model name is provided,
    or if the model name is not one of those that the user's API key has access to,
    then the user will be prompted for a model name.

    :param model_name: The name of the model to be used for default
    """
    from dotenv import load_dotenv

    from llamabot.config import llamabotrc_path

    load_dotenv(llamabotrc_path)

    client = OpenAI()

    model_list = client.models.list()
    available_models = [x.id for x in model_list if "gpt" in x.id]
    available_models.sort()

    if model_name in available_models:
        configure_environment_variable(
            env_var="DEFAULT_LANGUAGE_MODEL", env_value=model_name
        )
        return

    completer = WordCompleter(available_models)

    typer.echo("These are the GPT models available to you:")
    for model in available_models:
        typer.echo(model)

    while True:
        default_model = prompt(
            "Please type the name of the model you'd like to use: ",
            completer=completer,
            complete_while_typing=True,
            default=available_models[-1],
        )
        if default_model in available_models:
            configure_environment_variable(
                env_var="DEFAULT_LANGUAGE_MODEL", env_value=default_model
            )
            break



"""A bot designed to use questions and answers to generate a response."""
from hashlib import sha256
from pathlib import Path
from llamabot.bot.simplebot import SimpleBot
from llamabot.components.docstore import LanceDBDocStore
from llamabot.components.messages import AIMessage
from llamabot.doc_processor import magic_load_doc, split_document
from llamabot.prompt_manager import prompt
import json
import os
from tqdm.auto import tqdm

os.environ["TOKENIZERS_PARALLELISM"] = "false"


@prompt
def jeopardy_bot_sysprompt() -> str:
    """
    You are an expert at taking texts and constructing questions from them.
    You will be given a text.
    Extract as many question-and-answer pairs.
    Each answer may have multiple questions; be sure to cover as many as possible.
    Return a JSON array of the following schema:

    {
        "questions_and_answers": [
            {
                "question": "string",
                "answer": "string",
            },
            ...
        ]
    }
    """


@prompt
def q_and_a_prompt(query, q_and_a_results, relevant_documents) -> str:
    """Q&A Results:

        {{ q_and_a_results }}

    Relevant documents:

        {{ relevant_documents }}

    Query:

        {{ query }}
    """


class DocQABot:
    """A bot designed to use pre-computed questions and answers to generate a response."""

    def __init__(self, collection_name: str):
        self.question_store = LanceDBDocStore(
            collection_name=f"{collection_name}_questions"
        )
        self.document_store = LanceDBDocStore(collection_name=f"{collection_name}")

        self.response_bot = SimpleBot(
            "Based on Q&A results and relevant documents, please answer the query.",
            model_name="mistral/mistral-medium",
            api_key=os.getenv("MISTRAL_API_KEY"),
        )
        self.jeopardy_bot = SimpleBot(
            system_prompt=jeopardy_bot_sysprompt(),
            model_name="gpt-3.5-turbo-1106",
            json_mode=True,
            stream_target="stdout",
            temperature=0.3,
            api_key=os.getenv("OPENAI_API_KEY"),
        )

    def add_documents(
        self, document_paths: Path | list[Path], chunk_size=1_024, chunk_overlap=256
    ):
        """Add documents for the bot to query.

        :param document_paths: The paths to the documents to add.
        :param chunk_size: The size of the chunks to split the documents into.
        :param chunk_overlap: The amount of overlap between chunks.
        """
        if isinstance(document_paths, Path):
            document_paths = [document_paths]
        for document_path in tqdm(document_paths, desc="Document"):
            document = magic_load_doc(document_path)
            chunks = split_document(
                document, chunk_size=chunk_size, chunk_overlap=chunk_overlap
            )
            for chunk in tqdm(chunks, desc="Chunk Q&A"):
                doc_id = sha256(chunk.encode()).hexdigest()
                self.document_store.append(chunk, metadata=dict(doc_id=doc_id))
                q_and_a = json.loads(self.jeopardy_bot(chunk).content)
                for q_and_a in tqdm(
                    q_and_a["questions_and_answers"], desc="Storing Q&A"
                ):
                    q_a_concat = f"Q: {q_and_a['question']} A: {q_and_a['answer']}"
                    self.question_store.append(
                        q_a_concat, metadata=dict(parent_doc=doc_id)
                    )

    def __call__(
        self,
        query_text: str,
        num_questions_retrieved: int = 20,
        num_documents_retrieved: int = 3,
    ) -> AIMessage:
        """Call the QABot.

        This will retrieve the top `num_questions_retrieved` questions and answers
        from the question store, and the top `num_documents_retrieved` documents
        from the document store, and feed them into the response bot.

        :param query_text: The query text to use.
        :param num_questions_retrieved: The number of questions to retrieve.
        :param num_documents_retrieved: The number of documents to retrieve.
        :return: The response to the query, primed by the retrieved questions and
            documents.
        """
        # Retrieve from the question store
        q_and_as = self.question_store.retrieve(query_text)

        # Get the unique parent_doc IDs while preserving order
        result = self.question_store.collection.query(
            query_texts=query_text, n_results=num_questions_retrieved
        )
        parent_doc_ids = []
        for metadata in result["metadatas"][0]:
            if metadata["parent_doc"] not in parent_doc_ids:
                parent_doc_ids.append(metadata["parent_doc"])

        results = self.document_store.collection.query(
            query_texts=query_text,
            where={"doc_id": {"$in": list(parent_doc_ids)}},
            n_results=num_documents_retrieved,
        )

        output = self.response_bot(
            q_and_a_prompt(query_text, q_and_as, results["documents"][0])
        )
        return output
